//===================================RISCKER====================================
//==================================Data Path===================================
//------------------------------------------------------------------------------
//File Name:	datapath.v
//Purpose:		Data Path, main module that instantiates all RISCKER units
//Parameters:	None
//------------------------------------------------------------------------------
//=====================Alfredo Gualdron - Jose P. Pinilla=======================
//=================================Semillero ADT================================
//================================UPB Bucaramanga===============================
//Website:	http://semilleroadt.upbbga.edu.co/xiscker
//E-mail:	semilleroadt@gmail.com
//------------------------------------------------------------------------------
module datapath (
	input		clk,
	input		reset,
	input		[15:0] input_port,
	output	[15:0] output_port
	);

//////////////////////////////////////////////////////////////////
//======================Instructions Format=======================
//////////////////////////////////////////////////////////////////
wire [31:0] instruction; //IR Instruction Register

//R-Type
wire [3:0] opcode = 	instruction[31:28];
wire [5:0] rd = 		instruction[27:22];
wire [5:0] rs = 		instruction[21:16];
wire [5:0] rt = 		instruction[15:10];
wire [7:0] funct = 		instruction[7:0];

//I-Type
wire [15:0] imm = 				instruction[15:0];	 //Immediate

//J-Type
wire [15:0] target = 	instruction[15:0];

//////////////////////////////////////////////////////////////////
//=============Control Unit Finite State Machine==================
//////////////////////////////////////////////////////////////////
wire zero,sign;
wire [1:0] pc_src;
wire pc_next;
wire ir_write;
wire r2_src;				//Register 2 Address Source Select
wire rf_write;				// Register File Write Signal
wire alu_src;				//ALU Source Select
wire [10:0]operation;
wire alu_reg;				//ALU Result Register
wire mem_write; 			//Memory write pulse
wire [1:0] reg_src; 		//Register File Write Data Source Select
wire interruption;
wire retfi;
wire ipc_write;
wire intoggle;

reg rst = 0;
always@(negedge clk)
	rst <= reset;

FSM_Control controller (
    .fsm_clk(clk),
	 .reset(rst),
    .opcode(opcode), 
    .funct(funct), 
    .zero(zero), 
    .sign(sign), 
    .interruption(interruption),
    .retfi(retfi), 				
    .pc_src(pc_src), 
    .r2_src(r2_src), 
    .alu_src(alu_src), 
    .operation(operation), 
    .reg_src(reg_src), 
    .ipc_write(ipc_write), 		
    .intoggle(intoggle), 		
    .pc_next(pc_next), 
    .ir_write(ir_write), 
    .rf_write(rf_write), 
    .alu_reg(alu_reg), 
    .mem_write(mem_write)
    );

//////////////////////////////////////////////////////////////////
//===========================Input Port===========================
//////////////////////////////////////////////////////////////////

reg [15:0] ip=0; //Input Port 15 and 14 are connected as External Interruptions

always@(posedge ir_write)
	ip <= input_port;

//////////////////////////////////////////////////////////////////
//==============================Timers============================
//////////////////////////////////////////////////////////////////
wire [15:0] rc;

wire [1:0] prescaler_tr1 	=rc[9:8];
wire [1:0] prescaler_tr2 	=rc[11:10];
wire enable_tr1 			=rc[12];
wire enable_tr2 			=rc[13];
wire reset_tmr1				=rc[14];
wire reset_tmr2				=rc[15];
wire [15:0] tmr1;
wire [15:0] tmr2;
wire [15:0] ti1;
wire [15:0] ti2;
wire [15:0] eir1;
wire [15:0] eir2;
wire [15:0] pr1;
wire [15:0] pr2;

timer timer1 (
    .clk(clk), 
    .prescaler(prescaler_tr1), 
    .period(pr1), 
    .enable(enable_tr1), 
    .reset_tmr(reset_tmr1), 
    .int_timer(int_timer1), 
    .counter(tmr1)
    );
	 
timer timer2 (
    .clk(clk), 
    .prescaler(prescaler_tr2), 
    .period(pr2), 
    .enable(enable_tr2), 
    .reset_tmr(reset_tmr2), 
    .int_timer(int_timer2), 
    .counter(tmr2)
    );


//////////////////////////////////////////////////////////////////
//==================Interruptions Controller======================
//////////////////////////////////////////////////////////////////
wire [15:0] int_addr;

interrupt_controller int_controller (
    .clk(clk),
	 .intoggle(intoggle),
    .ipc_write(ipc_write),
    .int_ext1(ip[14]),
    .int_ext2(ip[15]),
    .int_timer1(int_timer1),
    .int_timer2(int_timer2),
    .rc(rc[7:0]),
    .ti1(ti1),
    .ti2(ti2),
    .eir1(eir1),
    .eir2(eir2),
    .interruption(interruption),
    .int_addr(int_addr),
    .retfi(retfi)
    );
	
//////////////////////////////////////////////////////////////////
//======================Program Counter===========================
//////////////////////////////////////////////////////////////////
wire [15:0] pc; 			
wire [15:0] rd1;			//Register File Data
wire [15:0] pc_plus1;

program_counter PC (
	 .reset(rst),
    .pc_next(pc_next),
    .pc_src(pc_src),
    .pc(pc),
    .target(target),
    .rd1(rd1),
	.int_addr(int_addr),
	.pc_plus1(pc_plus1)
    );

//////////////////////////////////////////////////////////////////
//=======Instruction Memory and IR(Instruction Register)==========
//////////////////////////////////////////////////////////////////
	
instmem	instmem_mif (
	.address ( pc ),
	.inclock ( ~clk ),
	.inclocken ( 1'b1 ),
	.outclock ( ir_write ),
	.q ( instruction )
	);
	
//////////////////////////////////////////////////////////////////
//=========================Register File==========================
//////////////////////////////////////////////////////////////////

wire [5:0] r2 = (r2_src) ? rd : rt;
wire [15:0] rd2;
reg [15:0] rf_data = 0;

regfile register_file (
    .rf_write(rf_write), 
    .intoggle(intoggle), 
    .ipc_write(ipc_write), 
    .r1(rs), 
    .r2(r2), 
    .wr(rd), 
    .wd(rf_data), 
    .ip(ip), 
    .tmr1(tmr1), 
    .tmr2(tmr2), 
    .ipc_in(pc),
    .rd1(rd1),
    .rd2(rd2), 
    .pr1(pr1), 
    .pr2(pr2), 
    .ti1(ti1), 
    .ti2(ti2), 
    .eir1(eir1), 
    .eir2(eir2), 
    .op(output_port), 
    .rc(rc)
    );

//////////////////////////////////////////////////////////////////
//==========================Operations============================
//////////////////////////////////////////////////////////////////
wire [15:0] b = alu_src ? imm : rd2;
wire [15:0] y; // ALU Result

operations alu_operations (
    .a(rd1), 
    .b(b), 
    .operation(operation), 
    .zero(zero), 
    .sign(sign),
    .y(y)
    );


//======ALU Register=======

reg [15:0] alu_result = 0;

always @ (posedge alu_reg)
alu_result <= y;

//////////////////////////////////////////////////////////////////
//=========================Data Memory============================
//////////////////////////////////////////////////////////////////
wire [15:0] write_data = rd2;
wire [15:0] mem_data;

dmem data_memory (
    .we(mem_write), 
    .addr(alu_result), 
    .wd(write_data), 
    .data(mem_data)
    );

//////////////////////////////////////////////////////////////////
//=================Register File Data Source======================
//////////////////////////////////////////////////////////////////
always @ (*)
	begin
		case(reg_src)
		2'd0: rf_data = alu_result;
		2'd1: rf_data = pc_plus1;
		2'd2: rf_data = mem_data;
		2'd3: rf_data = imm;
		endcase
	end

endmodule
